@@ -65,6 +65,7 @@ from hypothesis.internal.compat import (
     binary_type,
     get_type_hints,
     getfullargspec,
+    hbytes,
     int_from_bytes,
     qualname,
 )
@@ -92,10 +93,14 @@ from hypothesis.reporting import (
     verbose_report,
     with_reporter,
 )
-from hypothesis.searchstrategy.collections import TupleStrategy
-from hypothesis.searchstrategy.strategies import MappedSearchStrategy, SearchStrategy
 from hypothesis.statistics import note_engine_for_statistics
+from hypothesis.strategies._internal.collections import TupleStrategy
+from hypothesis.strategies._internal.strategies import (
+    MappedSearchStrategy,
+    SearchStrategy,
+)
 from hypothesis.utils.conventions import infer
+from hypothesis.vendor.pretty import CUnicodeIO, RepresentationPrinter
 from hypothesis.version import __version__
 
 if False:
@@ -309,13 +314,13 @@ class ArtificialDataForExample(ConjectureData):
         self.__draws = 0
         self.__kwargs = kwargs
 
-        def draw_bytes(data, n):
-            raise NotImplementedError()  # pragma: no cover
-
         super(ArtificialDataForExample, self).__init__(
-            max_length=0, draw_bytes=draw_bytes
+            max_length=0, prefix=hbytes(), parameter=None,
         )
 
+    def draw_bits(self, n):
+        raise NotImplementedError()  # pragma: no cover
+
     def draw(self, strategy):
         assert self.__draws == 0
         self.__draws += 1
@@ -503,6 +508,10 @@ class StateForActualGivenExecution(object):
 
         self.test = test
 
+        self.print_given_args = getattr(
+            wrapped_test, "_hypothesis_internal_print_given_args", True
+        )
+
         self.files_to_propagate = set()
         self.failed_normally = False
 
@@ -557,17 +566,57 @@ class StateForActualGivenExecution(object):
                         if expected_failure is not None:
                             text_repr[0] = arg_string(test, args, kwargs)
 
-                        if print_example:
-                            example = "%s(%s)" % (
-                                test.__name__,
-                                arg_string(test, args, kwargs),
-                            )
-                            report("Falsifying example: %s" % (example,))
-                        elif current_verbosity() >= Verbosity.verbose:
-                            report(
-                                lambda: "Trying example: %s(%s)"
-                                % (test.__name__, arg_string(test, args, kwargs))
-                            )
+                        if print_example or current_verbosity() >= Verbosity.verbose:
+                            output = CUnicodeIO()
+
+                            printer = RepresentationPrinter(output)
+                            if print_example:
+                                printer.text("Falsifying example:")
+                            else:
+                                printer.text("Trying example:")
+
+                            if self.print_given_args:
+                                printer.text(" ")
+                                printer.text(test.__name__)
+                                with printer.group(indent=4, open="(", close=""):
+                                    printer.break_()
+                                    for v in args:
+                                        printer.pretty(v)
+                                        # We add a comma unconditionally because
+                                        # generated arguments will always be
+                                        # kwargs, so there will always be more
+                                        # to come.
+                                        printer.text(",")
+                                        printer.breakable()
+
+                                    # We need to make sure to print these in the argument order for
+                                    # Python 2 and older versionf of Python 3.5. In modern versions
+                                    # this isn't an issue because kwargs is ordered.
+                                    arg_order = {
+                                        v: i
+                                        for i, v in enumerate(
+                                            getfullargspec(self.test).args
+                                        )
+                                    }
+                                    for i, (k, v) in enumerate(
+                                        sorted(
+                                            kwargs.items(),
+                                            key=lambda t: (
+                                                arg_order.get(t[0], float("inf")),
+                                                t[0],
+                                            ),
+                                        )
+                                    ):
+                                        printer.text(k)
+                                        printer.text("=")
+                                        printer.pretty(v)
+                                        printer.text(",")
+                                        if i + 1 < len(kwargs):
+                                            printer.breakable()
+                                printer.break_()
+                                printer.text(")")
+                            printer.flush()
+                            report(output.getvalue())
                         return test(*args, **kwargs)
 
         # Run the test function once, via the executor hook.
@@ -700,7 +749,7 @@ class StateForActualGivenExecution(object):
             return
         if runner.interesting_examples:
             self.falsifying_examples = sorted(
-                [d for d in runner.interesting_examples.values()],
+                runner.interesting_examples.values(),
                 key=lambda d: sort_key(d.buffer),
                 reverse=True,
             )
@@ -758,6 +807,12 @@ class StateForActualGivenExecution(object):
                 report("".join(traceback.format_exception(type(e), e, tb)))
 
             finally:  # pragma: no cover
+                # Mostly useful for ``find`` and ensuring that objects that
+                # hold on to a reference to ``data`` know that it's now been
+                # finished and they shouldn't attempt to draw more data from
+                # it.
+                ran_example.freeze()
+
                 # This section is in fact entirely covered by the tests in
                 # test_reproduce_failure, but it seems to trigger a lovely set
                 # of coverage bugs: The branches show up as uncovered (despite
